"""
uropa.py: UROPA - Universal RObust Peak Annotator

@authors: Annika Fust, Jens Preussner, Maria Kondili, Mette Bentsen and Mario Looso
@license: MIT
@maintainer: Mario Looso
@email: mario.looso@mpi-bn.mpg.de
"""

import os
import sys
import json
import argparse
import datetime
import time
import subprocess
import gzip
import copy
import psutil
import traceback

import logging
import multiprocessing as mp
import pysam

#Import internal functions
from .utils import *
from .annotation import *
from .__init__ import __version__ as VERSION

def restricted_float(f, f_min, f_max):
    f = float(f)
    if f < f_min or f > f_max:
        raise argparse.ArgumentTypeError("{0} not in range [0.0, 1.0]".format(f))
    return f

def split_options(options):
	#Splits ","-separated options from commandline into lists
	return sum([opt.split(",") if type(opt) == str else [opt] for opt in options], [])

def update_status(status_dict, task_list):
	""" Update status_dict with the information from task_list status """
	
	current_done = sum([task.ready() for task in task_list])
	status_dict["n_jobs_running"] = len(task_list) - current_done
	status_dict["n_jobs_done"] = status_dict["task_list_i"] + current_done #task_list_i is the previously finished jobs

	#Memory used
	status_dict["memory_percent"] = psutil.virtual_memory().percent

	return(status_dict)


def main():

	############################################################################################################
	#################################################### INPUT #################################################
	############################################################################################################

	start_time = datetime.datetime.now()
	cmd = " ".join(sys.argv)

	#----------------------------------------------------------------------------------------------------------#
	# Parse command-line arguments
	#----------------------------------------------------------------------------------------------------------#
	
	parser = argparse.ArgumentParser(
		prog="uropa",
		description='UROPA - Universal RObust Peak Annotator.',
		epilog=howtoconfig(),
		formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(prog, max_help_position=35, width=90))
	
	#Configuation arguments for one query
	one_query = parser.add_argument_group("Arguments for one query")
	one_query.add_argument("-b", "--bed", metavar="", help="Filename of .bed-file to annotate", action="store")
	one_query.add_argument("-g", "--gtf", metavar="", help="Filename of .gtf-file with features", action="store")
	one_query.add_argument("--feature", help="Feature for annotation", metavar="", nargs="*", default=[])

	one_query.add_argument("--feature-anchor", help="Specific feature anchor to annotate to", metavar="", choices=["start", "center", "end"], nargs="*", default=[])
	one_query.add_argument("--distance", help="Maximum permitted distance from feature (1 or 2 arguments)", metavar="", nargs="*", type=int, default=[1000,10000])
	one_query.add_argument("--strand", metavar="", help="Desired strand of annotated feature relative to peak", choices=['+', '-', 'same', 'opposite', 'ignore'], default='ignore')
	one_query.add_argument("--relative-location", metavar="", help="Peak location relative to feature location", nargs="*", choices=["PeakInsideFeature", "FeatureInsidePeak", "Upstream", "Downstream", "OverlapStart", "OverlapEnd"], default=[])
	one_query.add_argument("--internals", metavar="", help="Set minimum overlap fraction for internal feature annotations. 0 equates to internals=False and 1 equates to internals=True. Default is False.", type=lambda x: restricted_float(x, 0, 1), default=False)
	one_query.add_argument("--filter-attribute", metavar="", help="Filter on 9th column of GTF", default="")
	one_query.add_argument("--attribute-values", help="Value(s) of attribute corresponding to --filter-attribute", nargs="*", metavar="", default=[])
	one_query.add_argument("--show-attributes", help="A list of attributes to show in output (default: all)", metavar="", nargs="*", default=["all"])
	one_query.add_argument("--priority", help=argparse.SUPPRESS, action="store_true", default=False)

	#arguments for backwards compatibility using "-" instead of "_" in argument names
	one_query.add_argument("--relative_location", metavar="", help=argparse.SUPPRESS, nargs="*", choices=["PeakInsideFeature", "FeatureInsidePeak", "Upstream", "Downstream", "OverlapStart", "OverlapEnd"], default=[])	#deprecated but left for backwards compatibility
	one_query.add_argument("--feature_anchor", help=argparse.SUPPRESS, metavar="", choices=["start", "center", "end"], nargs="*", default=[])
	one_query.add_argument("--filter_attribute", metavar="", help=argparse.SUPPRESS, default="")	#deprecated but left for backwards compatibility
	one_query.add_argument("--attribute_values", help=argparse.SUPPRESS, nargs="*", metavar="", default=[])
	one_query.add_argument("--show_attributes", help=argparse.SUPPRESS, metavar="", nargs="*", default=[])

	#Or configuration arguments for multiple queries (overwrites)
	multi_query = parser.add_argument_group("Multi-query configuration file")
	multi_query.add_argument("-i", "--input", help="Filename of configuration file (keys in this file overwrite command-line arguments about query)", action="store", metavar="config.json")

	#Other arguments
	additional = parser.add_argument_group("Additional arguments")
	additional.add_argument("-p", "--prefix", metavar="", help="Prefix for result file names (defaults to basename of .bed-file)")
	additional.add_argument("-o", "--outdir", metavar="", help="Output directory for output files (default: current dir)", default=".")
	#additional.add_argument("-r","--reformat", help="create an additional compact and line-reduced table as result file", action="store_true")
	additional.add_argument("--output-by-query", help="Additionally create output files for each named query seperately", action="store_true")
	additional.add_argument("-s","--summary", help="Create additional visualisation of results in graphical format", action="store_true")
	additional.add_argument("-t","--threads", help="Multiprocessed run: n = number of threads to run annotation process", type=int, action="store", metavar="n", default=1)
	additional.add_argument("-l","--log", help="Log file name for messages and warnings (default: log is written to stdout)", action="store", metavar="uropa.log")
	additional.add_argument("-d","--debug",help="Print verbose messages (for debugging)", action="count", default=0)
	additional.add_argument("-v","--version", help="Prints the version and exits", action="version", version="%(prog)s " + VERSION)
	additional.add_argument("-c", "--chunk", metavar="", help="Number of lines per chunk for multiprocessing (default: 1000)", type=int, default=1000) 
	additional.add_argument("--target-mem", type=float, help=argparse.SUPPRESS, default=80)	#goal is to stay at maximum 80% memory consumption
	args = parser.parse_args()

	#Write help if no input was given
	if len(sys.argv[1:]) == 0:
		parser.print_help()
		sys.exit()

	#Check valid input to deal with the split between command-line query and multi-query config file
	if args.input == None:
		if args.bed == None:
			sys.exit("ERROR: --bed is needed for annotation without --input")
		if args.gtf == None:
			sys.exit("ERROR: --gtf is needed for annotation without --input")


	#----------------------------------------------------------------------------------------------------------#
	# Configure logger
	#----------------------------------------------------------------------------------------------------------#

	logger_q = mp.Manager().Queue()	#queue for multiprocessing logging
	logger = UROPALogger(debug_level=args.debug, log_f=args.log, q=logger_q) #debug_level is number of times --debug is given on commandline
	logger.start_logger_queue()	 #start listening for logging sent to queue
	logger_options = {"q": logger_q, "debug_level": args.debug}

	############################################################################################################
	############################################ VALIDATION OF INPUT ###########################################
	############################################################################################################
	
	logger.info("Started UROPA " + VERSION)
	logger.info("Working directory: {0}".format(os.getcwd()))
	logger.info("Command-line call: {0}".format(cmd))
	temp_files = []

	#----------------------------------------------------------------------------------------------------------#
	# Establish queries from command-line and --input
	#----------------------------------------------------------------------------------------------------------#

	logger.info("Reading configuration from commandline/input config")

	#First, fill in parameters from commandline
	cmd_query = {"feature":split_options(args.feature),
					 "feature_anchor":split_options(args.feature_anchor),
					 "distance": [args.distance[0], args.distance[0]] if len(args.distance) == 1 else args.distance,
					 "strand": args.strand,
					 "relative_location": split_options(args.relative_location),
					 "internals": args.internals,
					 "filter_attribute": args.filter_attribute,
					 "attribute_values": split_options(args.attribute_values),
					 }
	cmd_query["distance"] = split_options(cmd_query["distance"])

	valid_query_keys = set(list(cmd_query.keys()) + ["name"])

	#create cfg_dict like it would have been parsed from config .json
	cfg_dict = {"queries": [cmd_query],
				"show_attributes": split_options(args.show_attributes),
				"priority": args.priority,
				"gtf": args.gtf,
				"bed": args.bed,
				"prefix": args.prefix,
				"outdir": args.outdir,
				"threads": args.threads,
				"output_by_query": args.output_by_query
				}

	logger.debug("Config from command-line arguments: {0}".format(cfg_dict))

	#### Read from config file
	#Next, overwrite with config arguments if given, otherwise the arguments fall back to commandline default
	config = args.input
	if config != None:
		try:
			json_cfg_dict = parse_json(config)
			logger.debug("Config from json: {0}".format(json_cfg_dict))
			for key in json_cfg_dict:
				cfg_dict[key] = json_cfg_dict[key] 	#config values always win over commandline input
		except IOError:
			logger.error("File %s does not exists or is not readable.", config)
			sys.exit()
		except ValueError as e:
			logger.error("File %s contains malformed JSON. %s", config, e)
			sys.exit()

	#for key in json_cfg_dict:
		cfg_dict[key] = json_cfg_dict[key]

	# Validate output folder
	outdir = cfg_dict["outdir"]
	if not os.path.exists(outdir):
		try:
			logger.debug("Creating directory {}".format(outdir))
			os.makedirs(outdir)
		except Exception as e:
			logger.error(e)
			logger.error("Could not create directory {} for output".format(outdir))

	#Set prefix if not set
	if cfg_dict["prefix"] == None:
		cfg_dict["prefix"] = os.path.splitext(os.path.basename(cfg_dict["bed"]))[0]	
	output_prefix = os.path.join(cfg_dict["outdir"], cfg_dict["prefix"])
	logger.debug("Output_prefix set to: {0}".format(output_prefix))		

	#Format keys in cfg_dict and exit if error
	cfg_dict = format_config(cfg_dict, logger)
	logger.debug("Formatted config: {0}".format(cfg_dict))

	#Write out formatted config file
	cfg_dict_filled = copy.deepcopy(cfg_dict)
	for query in cfg_dict_filled["queries"]:
		query["relative_location"] = query.get("relative_location", ["PeakInsideFeature", "FeatureInsidePeak", "Upstream", "Downstream", "OverlapStart", "OverlapEnd"])
		query["strand"] = query.get("strand", 'ignore')
		query["feature_anchor"] = query.get("feature_anchor", ["start", "center", "end"])
	
	#----------------------------------------------------------------------------------------------------------#		
	# Validate existance of gtf / bed input and writability of output
	#----------------------------------------------------------------------------------------------------------#

	#Check if bed & gtf files exists
	for key in ["bed", "gtf"]:
		if cfg_dict[key] is not None:
			check_file_access(cfg_dict[key], logger)
		else:
			logger.error("No .{0}-file given as input - please check that a .{0}-file is given either via the commandline option --{0} or in the configuration file.".format(key))
			sys.exit()

	#Check whether output files can be written
	output_files = [os.path.join(output_prefix + suffix) for suffix in ["_allhits.txt", "_allhits.bed", "_finalhits.txt", "_finalhits.bed"]]
	for f in output_files:
		if os.path.exists(f) and not os.access(f, os.W_OK):
			logger.error("Output file {0} is not writable.".format(f))
			sys.exit()


	############################################################################################################
	################################################## PREPARATION #############################################
	############################################################################################################
		
	#----------------------------------------------------------------------------------------------------------#
	# Prepare GTF and extract chosen features to a subset gtf-file if needed
	#----------------------------------------------------------------------------------------------------------#		

	logger.info("Preparing .gtf-file for fast access")

	gtf_has_chr = check_chr(cfg_dict["gtf"])	#True/False

	#Check all possible features in gtf
	logger.debug("Finding all possible features in gtf")
	gtf_feat_count = {}
	with open(cfg_dict["gtf"]) as f:
		for line in f:
			if not line.startswith("#"):
				columns = line.rstrip().split("\t")
				
				if len(columns) < 9:
					logger.error("Input GTF ({0}) has less than 9 columns - please check that the file has the correct GTF format.".format(cfg_dict["gtf"]))
					sys.exit()

				feature = columns[2]
				if feature not in gtf_feat_count:
					gtf_feat_count[feature] = 0
				gtf_feat_count[feature] += 1

	gtf_feat = list(gtf_feat_count.keys())
	logger.debug("Features in gtf: {0}".format(gtf_feat_count))

	#Find all possible attributes in gtf
	logger.debug("Finding all possible attributes in gtf")
	gtf_attribute_count = {}
	with open(cfg_dict["gtf"]) as f:
		for line in f:
			if not line.startswith("#"):
				columns = line.rstrip().split("\t")
			
				pairs = columns[8].split(";")
				attributes = [pair.lstrip().rstrip().split(" ")[0] for pair in pairs]
				for att in attributes:
					gtf_attribute_count[att] = 0
	logger.debug("gtf attributes: {0}".format(list(gtf_attribute_count.keys())))

	if "all" in [str(att).lower() for att in cfg_dict.get("show_attributes", [])]:
		show_attributes = sorted(list(gtf_attribute_count.keys()))
		show_attributes = [att for att in show_attributes if att != ""]
		logger.info("Config key show_attributes was set to \'all\'. All possible attributes are shown in output ({0})".format(show_attributes))
	else:
		show_attributes = cfg_dict.get("show_attributes", [])

	#Fill in empty feature keys with all possible 
	for query in cfg_dict["queries"]:
		if len(query.get("feature", [])) == 0:
			query["feature"] = gtf_feat

	#Count all features given in config
	query_feat = []
	for query in cfg_dict["queries"]:
		query_feat.extend(query["feature"])
	query_feat = list(set(query_feat))

	#Check if any features were given that are not possible to subset on
	not_in_gtf = list(set(query_feat) - set(gtf_feat))
	if len(not_in_gtf) > 0:
		logger.error("Query feature(s) {0} not found in gtf".format(not_in_gtf))
		sys.exit()

	#Subset gtf if needed
	logger.debug("Subsetting gtf if needed")
	gtf_specific = list(set(gtf_feat) - set(query_feat)) 	#features in gtf which are not taken into account in queries
	if len(gtf_specific) > 0:
		sub_gtf = output_prefix + "_feature_subset.gtf"
		logger.debug("Subsetting {0} -> {1} with features {2}".format(cfg_dict["gtf"], sub_gtf, query_feat))
		subset_gtf(cfg_dict["gtf"], query_feat, sub_gtf)
		anno_gtf = sub_gtf
		temp_files.append(sub_gtf)
	else:
		anno_gtf = cfg_dict["gtf"]

	#Compress and index using gzip/tabix
	logger.debug("Tabix compress")
	anno_gtf_gz = output_prefix + ".gtf.gz"
	anno_gtf_index = anno_gtf_gz + ".tbi"

	success = 0
	sort_done = 0
	while success == 0:
		try:
			pysam.tabix_compress(anno_gtf, anno_gtf_gz, force=True)
			anno_gtf_gz = pysam.tabix_index(anno_gtf_gz, index=anno_gtf_index, seq_col=0, start_col=3, end_col=4, keep_original=True, force=True, meta_char='#')
			temp_files.extend([anno_gtf_gz, anno_gtf_index])
			success = 1
			if sort_done == 1:
				logger.info("Sorting and indexing was successful")

		except Exception as e:

			#Exit if we already tried to sort file once
			if sort_done == 1:
				logger.error("Could not index .gtf-file - please check whether the file has the correct 9-column format.")	
				sys.exit()
	
			#Read in and sort gtf file
			anno_gtf_sorted = output_prefix + "_sorted.gtf"
			temp_files.append(anno_gtf_sorted)
			sort_call = "grep -v \"^#\" {0} | sort -k1,1 -k4,4n > {1}".format(anno_gtf, anno_gtf_sorted)

			logger.warning("Indexing failed - the GTF is probably unsorted")
			logger.warning("Attempting to sort with call: {0}".format(sort_call))

			try:
				sub = subprocess.check_output(sort_call, shell=True)
			except subprocess.CalledProcessError:
				logger.error("Could not sort GTF file using command-line call: {0}".format(sort_call))
				sys.exit()

			anno_gtf = anno_gtf_sorted
			sort_done = 1

	#Write config used for annotation
	json_string = config_string(cfg_dict_filled)
	f = open(output_prefix + ".json", "w")
	f.write(json_string)
	f.close()

	############################################################################################################
	################################################## Annotation ##############################################
	############################################################################################################

	logger.info("Started annotation")
	prev_time = datetime.datetime.now()

	#Checking bed file format
	check_bed_format(cfg_dict["bed"], logger)

	#Setup pool for async annotation of peak chunks
	threads = int(cfg_dict["threads"])
	pool = mp.Pool(processes=threads, maxtasksperchild=100)
	task_list = []
	max_jobs_running = threads - 1
	chunk_size = args.chunk
	chunk_i = 0

	#Establish files to write
	file_dict = {key: os.path.join(output_prefix + "_" + key) for key in ["allhits.bed", "allhits.txt", "finalhits.bed", "finalhits.txt"]}

	#Additional files if output_by_query == True
	if cfg_dict["output_by_query"] == True:
		logger.info("Option --output-by-query is on. Writing additional hits-files per query.")
		query_names = [query["name"] for query in cfg_dict["queries"]]
		for name in query_names:
			file_dict[name + ".bed"] = os.path.join(output_prefix + "_" + get_valid_filename(name) + ".bed")
			file_dict[name + ".txt"] = os.path.join(output_prefix + "_" + get_valid_filename(name) + ".txt")
	
	logger.debug("File dictionary: {0}".format(file_dict))

	#Start queue for writing output
	manager = mp.Manager()
	q = manager.Queue()
	writer_task = pool.apply_async(sorted_file_writer, args=(q, file_dict, logger_options))

	##Initialize files with header line (or nothing)
	main = ["peak_chr", "peak_start", "peak_end", "peak_id", "peak_score", "peak_strand", "feature", "feat_start", "feat_end", "feat_strand", "feat_anchor", "distance", "relative_location", "feat_ovl_peak", "peak_ovl_feat"]
	header_output = main + show_attributes + ["name"]
	header_str = "\t".join(header_output) + "\n"
	for key in file_dict:
		if key.endswith(".txt"):
			q.put((key, 0, header_str))
		else:
			q.put((key, 0, ""))

	#Dict for collecting information on jobs, memory etc. during run
	status_dict = {"n_jobs_done": 0, 
				   "max_jobs_running": threads - 1, 
				   "n_jobs_running": 0,
				   "task_list_i": 0}

	#Go through bedfile and annotate chunks
	end_of_file = 0
	delta = datetime.timedelta(seconds=5)
	task_list_i = 0 #number of finished tasks
	with open(cfg_dict["bed"]) as f:

		while end_of_file == 0:

			#Get chunk of lines
			chunk_lines = [f.readline() for _ in range(chunk_size)]
			chunk_lines = [line for line in chunk_lines if line != ""]
			n_lines = len(chunk_lines)
			if n_lines == 0:
				break	#last chunk is empty; break out of while loop and don't annotate

			chunk_i += 1
	
			logger.debug("Read {0} lines from chunk {1} (({2}) - ({3}))".format(n_lines, chunk_i, " ".join(chunk_lines[0].split()[:3]), " ".join(chunk_lines[-1].split()[:3])))
			if n_lines < chunk_size:
				end_of_file = 1 #End of the file has been reached

			#Create peaks format for chunk
			peaks = parse_bedlines(chunk_lines, gtf_has_chr) #outputs list of peak-dictionaries

			if threads > 1:

				#Keep looping until job for this chunk has been started
				chunk_job_started = 0	
				while chunk_job_started == 0:

					#Check if writer is still runnning
					if writer_task.ready():
						logger.error("Writing of output files from multiprocessing jobs failed - please check any previous warnings.")
						sys.exit()

					#Update status dict
					status_dict = update_status(status_dict, task_list)

					#Write out progress after certain amount of time
					current_time = datetime.datetime.now()
					if current_time - prev_time > delta: 
						logger.info("Progress: Annotated {0} peaks ({1} jobs running; {2} jobs finished)".format(status_dict["n_jobs_done"] * chunk_size, 
																														     status_dict["n_jobs_running"], 
																														     status_dict["n_jobs_done"]))
						logger.debug("Max jobs allowed: {0}; Virtual memory percent: {1}; Writer task running: {2}; Writer queue size: {3}; logger queue size: {4}".format(status_dict["max_jobs_running"], 
																																											status_dict["memory_percent"], 
																																											not writer_task.ready(), 
																																											q.qsize(),
																																											logger_q.qsize()))
						logger.debug("Tasks in cache: {0}".format(list(pool._cache.keys())))
						prev_time = current_time

					#Test memory consumption and adjust max_jobs_running accordingly
					if status_dict["memory_percent"] == 100:
						logger.warning("Memory usage has reached 100% - this might cause UROPA to halt or crash. Reducing '--chunk' or '--threads' can help to resolve this issue.")	

					#How many jobs should be running to fulfill --target-mem consumption?
					memory_per_job = status_dict["memory_percent"] / max(1, status_dict["n_jobs_running"])
					status_dict["max_jobs_running"] = min(threads - 1, int(np.floor(args.target_mem / memory_per_job))) #maximum possible is threads - 1, but can be lower if memory becomes low
					status_dict["max_jobs_running"] = max(max_jobs_running, 1) # minimum 1 job running

					#### Check whether new jobs can be send to pool ###
					if status_dict["n_jobs_running"] >= status_dict["max_jobs_running"]:
						time.sleep(0.5) #wait half a second, and continue while-loop (if condition is True)

					else:
						#Add new chunk to pool
						logger.debug("Adding job for chunk {0}".format(chunk_i))
						task_list.append(pool.apply_async(annotate_peaks, args=(peaks, anno_gtf_gz, anno_gtf_index, cfg_dict, q, chunk_i, show_attributes, logger_options)))
						chunk_job_started = 1

					#Check exit status of any finished tasks
					for i, task in enumerate(task_list):
						if task.ready():
							try:
								task.get() #either successful or exited with exception
							except Exception as e:
								logger.error("Multiprocessing task exited with error: {0}".format(type(e)))
								logger.error("Please check previous warnings for resolving the issue.")
								logger.error("Full traceback:\n{0}".format(traceback.format_exc()))
								sys.exit(1)

					#If the first task in task_list was finished, it can be removed to reduce size of task_list
					while len(task_list) > 0 and task_list[0].ready():
						del task_list[0]
						status_dict["task_list_i"] += 1

			else:
				logger.debug("Annotating peak chunk")
				annotate_peaks(peaks, anno_gtf_gz, anno_gtf_index, cfg_dict, q, chunk_i, show_attributes, logger_options)

	# Whole file has been read and all jobs were added to task_list
	# Wait for all jobs to finish
	logger.info("The input .bed-file has been read and all jobs have been started. Waiting for the final tasks to finish...")
	while sum([task.ready() for task in task_list]) < len(task_list):
		logger.debug("Tasks in cache: {0}".format(list(pool._cache.keys())))
		time.sleep(0.5) 

	#Check results for final tasks
	for task in task_list:
		try:
			task.get() #either successful or exited with exception
		except Exception as e:
			logger.error("Multiprocessing task exited with error: {0}".format(type(e)))
			logger.error("Please check previous warnings for resolving the issue.")
			logger.debug("Full traceback:\n{0}".format(traceback.format_exc()))
			sys.exit(1)

	#End writer queue
	q.put((None, None, None))

	#Check that all queues are done writing
	while q.qsize() != 0:
		logger.debug("- Queue size {0}".format(q.qsize()))
		time.sleep(0.5) 

	writer_result = writer_task.get()	#locks until writer_task returns value (i.e. files were closed)

	pool.terminate()
	pool.join()

	############################################################################################################
	################################################ POSTPROCESSING ############################################
	############################################################################################################

	logger.info("Processing annotated peaks")

	"""
	##### Check if no annotations were found #####
	all_NA = 0
	for anno in all_annotations:
		if "feature" in anno:
			all_NA = 1
	if all_NA == 0:	#This is 0 coming out of the loop if no features were found
		logger.warning("No annotations were found for input regions (all hits are NA). If this is unexpected, please check the configuration of your input queries.")
	"""

	##### Visual summary #####
	if args.summary:
		logger.info("Creating the Summary graphs of the results...")
		summary_script = "uropa_summary.R"
		summary_output = output_prefix + "_summary.pdf"

		#cmd is the command-line call str
		call = [summary_script, "-f", os.path.join(output_prefix + "_finalhits.txt"), "-c", output_prefix + ".json", "-o", summary_output, "-b", os.path.join(output_prefix + "_allhits.txt"), "-a \'", cmd, "\'"]
		call_str = ' '.join(call)
		
		try:
			logger.debug('Summary output call is {}'.format(call_str))
			sum_pr = subprocess.check_output(call_str, shell=True)
		except subprocess.CalledProcessError:
			logger.warning("Visualized summary output could not be created from: %s", call_str)
		except OSError:
			logger.warning("Rscript command not available for summary output.")

	##### Cleanup #####
	if args.debug == False:
		logger.info("Cleaning up temporary files")
		for f in temp_files:
			try:
				os.remove(f)
			except:
				logger.warning("Could not remove temporary file {0}".format(f))

	end_time = datetime.datetime.now()
	total_time = end_time - start_time
	logger.info("UROPA run finished in {0}!".format(str(total_time).split('.', 2)[0]))

	logger.stop_logger_queue()	#done logging



"""Contains code for the annotation process"""

import pysam
from functools import reduce
import re
import numpy as np
import logging
import datetime

from .utils import * #import Uropa logging functions

def decimal_round(num, d=3):

	stringnum = "{:.20f}".format(num)

	#Position of first non-zero integer after decimal
	decimal_i = stringnum.index(".")
	nonzero_i_lst = [i for i in range(decimal_i+1, len(stringnum)) if int(stringnum[i]) > 0] + [0]	#If all describing digits are 0 (e.g. 0.0 or 1.0), nonzero_is is [0]
	nonzero_i = nonzero_i_lst[0]

	#Decide how to round
	if nonzero_i < decimal_i + d:	#If a non-zero integer was already found
		rounded = str(round(num, d))

	elif nonzero_i >= decimal_i + d: #If a non-zero integer was already found
		rounded = stringnum[:nonzero_i+1]

	return(rounded)

def create_anno_dict(peak, hit):
	""" Returns a dictionary containing information on the hit from gtf """

	#Add peak information
	anno_dict = {}
	anno_dict.update(peak) #fills out peak chr/start/end/id/score/strand
	anno_dict["peak_center"] = int((anno_dict["peak_end"] + anno_dict["peak_start"])/2.0)
	anno_dict["peak_length"] = anno_dict["peak_end"] - anno_dict["peak_start"]

	#Parse info from gtf string
	try:
		att = hit.attributes.strip("; ")	#remove trailing ;
		pairs = re.split(";\s*", att) 		#regex split on ;(space)

		pairs = [pair.split() for pair in pairs]
		attribute_tags = [pair[0] for pair in pairs]
		attribute_values = [" ".join(pair[1:]).replace("\"", "") for pair in pairs]
		
		#Join values from identical tags to list
		attribute_dict = {tag:[] for tag in attribute_tags}
		for i, (tag, value) in enumerate(zip(attribute_tags, attribute_values)):
			attribute_dict[tag].append(value)

	except Exception as e:
		print("Error reading attributes: {0} ({1})".format(hit.attributes, e))
		attribute_dict = {}

	# Fill in with feature info
	anno_dict["feature"] = hit.feature
	anno_dict["feat_strand"] = hit.strand
	anno_dict["feat_start"] = int(hit.start)
	anno_dict["feat_end"] = int(hit.end)
	anno_dict["feat_center"] = int((anno_dict["feat_end"] + anno_dict["feat_start"])/2.0)
	anno_dict["feat_length"] =  int(anno_dict["feat_end"] - anno_dict["feat_start"])
	anno_dict["feat_attributes"] = attribute_dict	# Dictionary of lists containig values {"key1":[value1], "key2":[val1,val2]} 

	#Look-up keys for annotation
	anno_dict["anchor_pos"] = {"start": anno_dict["feat_start"] if anno_dict["feat_strand"] != "-" else anno_dict["feat_end"],
								"center": anno_dict["feat_center"], 
								"end": anno_dict["feat_end"] if anno_dict["feat_strand"] != "-" else anno_dict["feat_start"]}

	#Change with each query
	anno_dict["query"] = 0	 #query for which the hit was valid
	anno_dict["best_hit"] = 0

	return(anno_dict)



def distance_to_peak_center(anno_dict, query_anchor):
    """ Assigns the distance of peak center to best query anchor """

    anchor_list = list(query_anchor)

    #Set default if anchor list is empty
    if len(query_anchor) == 0:
    	anchor_list = ["start", "center", "end"]

    #Calculate distances to each possible anchor
    raw_distances = [anno_dict["peak_center"] - anno_dict["anchor_pos"][anchor] for anchor in anchor_list]
    abs_distances = [abs(dist) for dist in raw_distances]
    min_dist_i = abs_distances.index(min(abs_distances))

    #Set minimum distance as best anchor
    anno_dict["raw_distance"] = raw_distances[min_dist_i]
    anno_dict["distance"] = int(abs(raw_distances[min_dist_i]))
    anno_dict["feat_anchor"] = anchor_list[min_dist_i]

    return(anno_dict)


# import "division" allows decimals
def calculate_overlap(anno_dict):
    """ Calculates percentage of length covered by the peak/feature """
    
    #beds exclude first position, therefore +1 for starts. Range excludes last position in range - therefore +1 for end
    ovl_range = range(max(anno_dict["peak_start"]+1, anno_dict["feat_start"]+1), min(anno_dict["peak_end"], anno_dict["feat_end"])+1)
    ovl_bp = len(ovl_range)

    ovl_pk = ovl_bp / float(anno_dict["peak_length"])
    ovl_feat = ovl_bp / float(anno_dict["feat_length"])

    anno_dict["feat_ovl_peak"] = decimal_round(ovl_pk, 3)    #Percentage of the peak that is covered by the feature (1.0 corresponds to the genomic_location “PeakInsideFeature”). 
    anno_dict["peak_ovl_feat"] = decimal_round(ovl_feat, 3)	 #Percentage of the feature that is covered by the peak (1.0 corresponds to the genomic_location “FeatureInsidePeak”).

    return(anno_dict)


def get_relative_location(anno_dict):
	""" Sets the relative location of peak to feature """

	location = "NA"
	if float(anno_dict["feat_ovl_peak"]) == 1:
		location = "PeakInsideFeature"

	elif float(anno_dict["peak_ovl_feat"]) == 1:
		location = "FeatureInsidePeak"

	elif float(anno_dict["feat_ovl_peak"]) == 0: #no overlap
		
		#If feature (gene) is on the - strand
		if anno_dict["feat_strand"] == "-":
			
			#Check if peak is upstream/downstream of anchor
			if anno_dict["peak_center"] > anno_dict["anchor_pos"][anno_dict["feat_anchor"]]:
				location = "Upstream" #For feature strand '-': If peak_center is higher than the anchor, the peak is upstream
			else:
				location = "Downstream"

		#Else, feature must be on the plus strand
		else:
			#Check if peak is upstream/downstream of anchor
			if anno_dict["peak_center"] > anno_dict["anchor_pos"][anno_dict["feat_anchor"]]:
				location = "Downstream" #For feature strand '+': If peak_center is higher than the anchor, the peak is downstream
			else:
				location = "Upstream"
	
	else:	#some overlap, but not completely internal

		#Figure out if it is overlapstart or overlap end
		if anno_dict["anchor_pos"]["start"] in range(anno_dict["peak_start"]+1, anno_dict["peak_end"]+1):
			location = "OverlapStart"

		elif anno_dict["anchor_pos"]["end"] in range(anno_dict["peak_start"]+1, anno_dict["peak_end"]+1):
			location = "OverlapEnd"

	anno_dict["relative_location"] = location

	return(anno_dict)


def annotate_single_peak(peak, tabix_obj, cfg_dict, logger=None):
	""" 
	Annotates a single peak against tabix_obj (GTF) using the configuration in queries

	Input:
		peak (dict): A dictionary containing information on peak. It has the form: {"peak_chr":"chr1", "peak_start":0, "peak_end":100, "peak_id":"peak1", "peak_score":0.4, "peak_strand":"+"}
					 This dict can also include gtf_chr in case the gtf has a different chromosome prefix, e.g.: { (...), "gtf_chr":1}
					 It might also contain other information such as: { (...), "additional_bed_column_1":"column_value"}
		tabix_obj (Pysam TabixFile): tabix_obj is an open pysam.TabixFile
		cfg_dict (dict): Configuration dictionary containing "queries" and "priority" keys
		logger (python.logging obj): Logger object for debugging or None

	Returns:
		valid_annotations (list): A list of dictionaries containing the full peak dictionary filled in with additional keys related to the annotated element. 
								  If no hits were found, the peak is returned without annotation information.
	"""

	#General info on queries
	queries = cfg_dict["queries"]
	n_queries = len(queries)
	distances = sum([query["distance"] for query in queries], [])
	max_distance = int(max(distances))

	#Annotation
	logger.debug2("Annotating peak: {0}".format(peak))

	valid_annotations = []	#for this peak
	stop_searching = False
	query_i = -1
	while query_i+1 < n_queries and stop_searching == False:

		query_i += 1				#First query is 0
		query = queries[query_i]	#current query to check

		logger.debug2("Finding hits for query: {0}".format(query))

		#Extend and fetch possible hits from tabix
		max_distance = max(query["distance"])
		extend_start = int(max(1, peak["peak_start"] - max_distance))
		extend_end = peak["peak_end"] + max_distance
		tabix_query = "{0}:{1}-{2}".format(peak.get("gtf_chr", peak["peak_chr"]), extend_start, extend_end)
		logger.debug2("Tabix query for query {0} ({1}): {2}".format(query_i, query["name"], tabix_query))

		try:
			begin = datetime.datetime.now()
			hits = list(tabix_obj.fetch(tabix_query, parser=pysam.asGTF()))	#hits for this query
			end = datetime.datetime.now()
			logger.debug2("Fetched {0} hits in {1}".format(len(hits), end - begin))
		except: 
			#exception if no hits could be fetched from tabix, for example if the contig does not exist in the gtf index.
			#print("ERROR: Could not create iterator for peak: {0} \n with tabix query {1}".format(peak, tabix_query))		
			stop_searching = True 
			hits = []

		begin = datetime.datetime.now()
		for hit in hits: 
			
			#If feature is not the right one, we do not have to go further - saves computation of distances
			if "feature" in query:
				if hit.feature not in query["feature"]:
					#logger.debug("{0} not in {1} - continuing to next hit".format(hit.feature, query["feature"]))
					continue

			anno_dict = create_anno_dict(peak, hit)
			anno_dict["query"] = query_i
			anno_dict["query_name"] = query["name"]

			#Calculate distances/relative location
			anno_dict = distance_to_peak_center(anno_dict, query.get("feature_anchor", []))
			anno_dict = calculate_overlap(anno_dict)
			anno_dict = get_relative_location(anno_dict)

			##### Test validity of the hit to query_i ####
			checks = {}
			#if "feature" in query:
			#	checks["feature"] = anno_dict["feature"] in query["feature"]	 #feature

			#Check feature anchor
			if "feature_anchor" in query:
				checks["feature_anchor"] = anno_dict["feat_anchor"] in query["feature_anchor"]

			#Peak strand relative to feature strand
			if "strand" in query:
				if query["strand"] != "ignore" and anno_dict["peak_strand"] != ".":
					checks["strand"] = ((anno_dict["peak_strand"] == query["strand"]) or
										(anno_dict["peak_strand"] == anno_dict["feat_strand"] and query["strand"] == "same") or 
										(anno_dict["peak_strand"] != anno_dict["feat_strand"] and query["strand"] == "opposite"))	

			#Check whether distance was valid
			if anno_dict["feat_strand"] == "+":
				checks["distance"] = anno_dict["raw_distance"] > -query["distance"][0] and anno_dict["raw_distance"] < query["distance"][1]
			else:
				checks["distance"] = anno_dict["raw_distance"] > -query["distance"][1] and anno_dict["raw_distance"] < query["distance"][0]

			#Check distance (Distance can still be valid if PeakInsideFeature/FeatureInsidePeak and internals flag is set)
			if "internals" in query:
				max_overlap = max(float(anno_dict["feat_ovl_peak"]), float(anno_dict["peak_ovl_feat"]))
				checks["distance"] = checks["distance"] or (query["internals"]*1.0 > 0 and max_overlap >= query["internals"]*1.0)	#if internals is set to more than 0 overlap

			#Filter on relative location
			if "relative_location" in query:
				checks["relative_location"] = anno_dict["relative_location"] in query["relative_location"]
			
			#Filter on attribute if any was set
			if "filter_attribute" in query and "attribute_values" in query: #query["filter_attribute"] != None:

				#Check if the desired filter attribute is in the attributes of the hit:
				checks["attribute"] = False
				if query["filter_attribute"] in anno_dict["feat_attributes"]:
					for filter_value in query["attribute_values"]:
						tag_values_list = anno_dict["feat_attributes"][query["filter_attribute"]]	#list of values for this tag
						if filter_value in tag_values_list:
							checks["attribute"] = True
			
			##### All checks are done -> establish if hit is a valid annotation #####
			valid = sum(checks.values()) == len(checks.values()) #all checks must be valid
			if valid:
				valid_annotations.append(anno_dict.copy())
			logger.debug2("Validity: {0} | Checks: {1} | Annotation dict: {2}".format(valid, checks, {key:anno_dict[key] for key in anno_dict}))
		
		end = datetime.datetime.now()
		logger.debug2("Validated hits in {0}".format(end-begin))

		#All tabix hits for this query were checked - if priority, stop searching if any valid hit was found -> else, check next query
		stop_searching = (len(valid_annotations) > 0 and cfg_dict["priority"]) or stop_searching #or if stop_searching was already set previously

		if stop_searching == True:
			logger.debug2("{0} valid hit(s) were found for query_i = {1} and priority is true - stopping search.".format(len(valid_annotations), query_i))
		else:
			logger.debug2("A total of {0} valid hits were found. Incrementing query_i.".format(len(valid_annotations), query_i))
	
	logger.debug2("")	#create empty line in debugger output for easy overview

	#After all hits have been checked for peak; make final checks and set best flag
	if len(valid_annotations) > 0:	

		#Sort valid annotations
		valid_annotations = sorted(valid_annotations, key= lambda anno_dict: (anno_dict["feat_start"], anno_dict["feat_end"], anno_dict["feature"]))
		
		#If priority == True, find the highest ranked annotations for this peak
		#if cfg_dict["priority"] == True:
		#	highest_priority_query = min([anno_dict["query"] for anno_dict in valid_annotations])
		#	#logger.debug("Highest priority hit for {0} was {1}".format(peak, highest_priority_query))
		#	valid_annotations = [anno_dict for anno_dict in valid_annotations if anno_dict["query"] == highest_priority_query]

		distances = [anno_dict["distance"] for anno_dict in valid_annotations]
		best_hit_index = distances.index(min(distances))
		valid_annotations[best_hit_index]["best_hit"] = 1

	else:
		valid_annotations = [peak]
		valid_annotations[0]["best_hit"] = 1			#the empty hit for the peak is the best hit

	return(valid_annotations)


def annopeak_to_string(annotated_peak, columns=None, attributes=[]):
	""" Format an annotated peak to a .bed-file output """

	#Default output
	if columns == None:
		columns = ["peak_chr", "peak_start", "peak_end", "peak_id", "peak_score", "peak_strand"]
		columns += ["feature", "feat_start", "feat_end", "feat_strand", "feat_anchor", "distance", "relative_location", "feat_ovl_peak", "peak_ovl_feat"]

	#Fetch columns from peak in the correct order
	annotation_columns = [str(annotated_peak.get(key, "NA")) for key in columns]
	attribute_columns = [",".join(annotated_peak.get("feat_attributes", {}).get(key, ["NA"])) for key in attributes]

	bed_str = "\t".join(annotation_columns + attribute_columns + [annotated_peak.get("query_name", "NA")])

	return(bed_str)


def annotate_peaks(peaks, gtf_gz, gtf_index, cfg_dict, q, idx, attributes, logger_options):
	""" 
		Input: 
		peaks (list): List of dictionaries containing information on peaks to annotate (see function 'annotate_single_peak')
		gtf_gz (str): Path to gtf.gz file
		gtf_index (str): Path to gtf.gz index 
		cfg-dict (dict): The loaded config containing queries
		q (Queue): The queue to put annotations into
		idx (int): The order in which the annotations should be written to output
		attributes (list): A list of attribute columns to write to output,
		logger_options (dict): A dict for initializing UROPALogger
	"""

	logger = UROPALogger(**logger_options)

	#Open tabix file
	tabix_obj = pysam.TabixFile(gtf_gz, index=gtf_index)

	#For each peak in input peaks, collect all_valid_annotations
	logger.debug("Annotating peaks in chunk {0}".format(idx))
	all_valid_annotations = []
	for peak in peaks:
		
		#Annotate single peak
		valid_annotations = annotate_single_peak(peak, tabix_obj, cfg_dict, logger=logger)
		all_valid_annotations.extend(valid_annotations)

	tabix_obj.close()

	#Write annotations to best hits and final hits
	logger.debug("Annotated all peaks in chunk {0}. Now adding contents to queue...".format(idx))
	content = "\n".join([annopeak_to_string(peak, attributes=attributes) for peak in all_valid_annotations]) + "\n"
	q.put(("allhits.bed", idx, content))
	q.put(("allhits.txt", idx, content))
	content = ""
	
	finalhits_content = "\n".join([annopeak_to_string(peak, attributes=attributes) for peak in all_valid_annotations if peak.get("best_hit", 0) == 1]) + "\n"
	q.put(("finalhits.bed", idx, finalhits_content))
	q.put(("finalhits.txt", idx, finalhits_content))
	finalhits_content = ""
	
	## Hits per query if chosen
	if cfg_dict["output_by_query"] == True:
		query_names = [query["name"] for query in cfg_dict["queries"]]
		for name in query_names:
			query_str = "\n".join([annopeak_to_string(peak, attributes=attributes) for peak in all_valid_annotations if peak.get("query_name", "") == name]) + "\n"
			q.put((name + ".bed", idx, query_str))
			q.put((name + ".txt", idx, query_str))

	logger.debug("Job finished for chunk {0}".format(idx))
	return(0) #success




"""Contains helper functions for UROPA """
import json
from textwrap import dedent
import ast
import numpy as np
import os
import sys
import re
import traceback
import logging
from logging import handlers
import multiprocessing as mp
import time
import datetime

class UROPALogger(logging.Logger):

	logger_levels = {
						0: logging.INFO,
						1: logging.DEBUG,		#debugging info
						2: logging.DEBUG - 5	#deeper debugging info
					}

	def __init__(self, debug_level=0, log_f=None, q=None):
		""" 
		debug_level: Logging level as identified in logger_levels
		log_f: path to output logfile (in addition to stdout)
		q: queue for handling logging via QueueHandler 
		"""

		self.debug_level = debug_level
		self.log_f = log_f
		self.q = q

		#Create logger
		logging.Logger.__init__(self, "UROPA")

		####### Setup custom levels #######
		#Create custom level for deeper debug messages
		deeper_level = UROPALogger.logger_levels[2]
		logging.addLevelName(deeper_level, "DEBUG2")
		setattr(self, 'debug2', lambda *args: self.log(deeper_level, *args))

		#Set level and format
		self.level = UROPALogger.logger_levels[debug_level]
		self.setLevel(self.level)
		self.formatter = logging.Formatter("%(asctime)s (%(process)d) [%(levelname)s]\t%(message)s", "%Y-%m-%d %H:%M:%S")

		########## Setup streaming #########
		#Log file stream
		if log_f != None:
			try:
				log = logging.FileHandler(log_f, "w")
				log.setLevel(self.level)
				log.setFormatter(self.formatter)
				self.addHandler(log)
			except:
				sys.exit("ERROR: Could not create logfile '{0}'. Please check that the given path exists.".format(log_f))

		#Stdout stream
		if self.q == None:
			con = logging.StreamHandler(sys.stdout)		#console output
			con.setLevel(self.level)
			con.setFormatter(self.formatter)
			self.addHandler(con)

		#Stdout handled by queue
		else:
			h = logging.handlers.QueueHandler(self.q)  	# Just the one handler needed
			self.handlers = []
			self.addHandler(h)

		#logger is ready for use

	def start_logger_queue(self):
		""" start process for listening and handling through the main logger queue """

		self.listener = mp.Process(target=self.main_logger_process)
		self.listener.start()

	def main_logger_process(self):
		""" This process runs in the main thread and handles the logging events """

		mainlogger = UROPALogger(self.debug_level, self.log_f)	#no queue, this logger is for collecting logging from the queue
		mainlogger.debug("Started main logger process")
		while True:
			try:
				record = self.q.get()
				if record is None:
					break
				mainlogger.handle(record) 	#this logger is running in the main process

			except EOFError:
				mainlogger.error("Logger lost connection to queue - probably due to an error raised from a child process.")
				break

		return(0)

	def stop_logger_queue(self):
		""" Stop process for listening """

		self.debug("Waiting for listener to finish")
		self.q.put(None)
		while self.listener.exitcode != 0:
			self.debug("Listener exitcode is: {0}. Waiting for exitcode = 0.".format(self.listener.exitcode))
			time.sleep(0.1)

		self.debug("Joining listener")
		self.listener.join()
	

#Validation of config
def howtoconfig():
    """Defines the epilog that is given when help is requested."""
    epilog = dedent("""\
    UROPA is a peak annotation tool facilitating the analysis of next-generation sequencing methods such
    as ChIPseq and ATACseq. The advantage of UROPA is that it can accommodate advanced structures of annotation
    requirements. UROPA is developed as an open source analysis pipeline for peaks generated from standard peak callers.

    Please cite upon usage:
    Kondili M, Fust A, Preussner J, Kuenne C, Braun T and Looso M. UROPA: A tool for Universal RObust Peak Annotation.
    Scientific Reports 7 (2017), doi: 10.1038/s41598-017-02464-y

    Please visit http://uropa-manual.readthedocs.io/config.html for detailed information on configuration.

    """)
    """
    More information on how to set up the config-file is found at:
    https://uropa-manual.readthedocs.io/config.html


    The configuration file should at least contain paths for bed and GTF files:

    {
    "queries": [],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf"
    }

    Different query types can be defined using the queries key:

    {
    "queries": [
      {...},
      {...}],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf"
    }

    Optionally, the priority key can be used to fine tune UROPAs behaviour:

    {
    "queries": [
      {...},
      {...}],
    "bed": "/path/to/bed/file.bed",
    "gtf": "/path/to/annotation/file.gtf",
    "priority": "True"
    }

    Please visit http://uropa-manual.readthedocs.io/config.html for detailed information on configuration.
	"""
    return epilog

def get_valid_filename(s):
	"""
	Return the given string converted to a string that can be used for a clean
	filename. Remove leading and trailing spaces; convert other spaces to
	underscores; and remove anything that is not an alphanumeric, dash, underscore, or dot.
	from django.utils.text
	"""
	s = str(s).strip().replace(' ', '_')
	return(re.sub(r'(?u)[^-\w.]', '', s))

def parse_json(infile):
    """ Read a json file to a dict """
    with open(infile) as f:
        return ast.literal_eval(json.dumps(json.load(f)))


def format_config(cfg_dict, logger):
	""" Format input config to catch any errors """
	
	query_prefix = "query_"
	case_conversion = {"peakinsidefeature": "PeakInsideFeature", "featureinsidepeak": "FeatureInsidePeak", 
						"upstream": "Upstream", "downstream":"Downstream", "overlapstart":"OverlapStart", "overlapend": "OverlapEnd"}
	convert_values = {value: True for value in [True, 't', 'true', 'yes', 'y']}
	convert_values.update({value: False for value in [False, 'f', 'false', 'no', 'n']})
	

	###################### Upper level config keys ####################
	#Check existance of upper-level keys
	valid_config_keys = set(["queries", "priority", "show_attributes", "gtf", "bed", "prefix", "outdir", "threads", "output_by_query"])
	given_config_keys = set(cfg_dict.keys())
	invalid_config_keys = given_config_keys - valid_config_keys
	if len(invalid_config_keys) > 0:
		logger.error("Error in config file: Key(s) {0} not known to UROPA.".format(invalid_config_keys))
		sys.exit()

	#Delete empty keys
	for key in list(cfg_dict.keys()):
		if is_empty(cfg_dict[key]):
			del cfg_dict[key]

	#Check format of values
	if "show_attributes" not in cfg_dict:
		cfg_dict["show_attributes"] = []
	else:
		if not isinstance(cfg_dict["show_attributes"], list):	#make sure that show_attributes is a list
			cfg_dict["show_attributes"] = [cfg_dict["show_attributes"]]

	#Convert all forms of true/false to python bool value
	cfg_dict["priority"] = convert_values[str(cfg_dict.get("priority", False)).lower()] 	#per default false
	cfg_dict["output_by_query"] =  convert_values[str(cfg_dict.get("output_by_query", False)).lower()]

	####################### Query level keys ##########################
	#Convert keys for backwards compatibility {old:new, (...)}
	conversion = {"show.attributes":"show_attributes", "feature.anchor": "feature_anchor", "filter.attribute":"filter_attribute", 
					"attribute.values": "attribute_values", "attribute.value": "attribute_values", "attribute_value": "attribute_values", 
					"direction": "relative_location"}
	valid_query_keys = set(["feature", "feature_anchor", "distance", "strand", "relative_location", "internals", "filter_attribute", "attribute_values", "name"])

	#Check and fill queries with needed defaults such as distance
	for i, query in enumerate(cfg_dict["queries"]):

		#Convert keys for backwards compatibility
		for key in list(query.keys()):
			if key in conversion:
				query[conversion[key]] = query[key]
				del query[key]

		#Move show_attributes to upper_level if given
		if "show_attributes" in query:
			if type(query["show_attributes"]) == list:
				for att in query["show_attributes"]:
					if att not in cfg_dict["show_attributes"]:
						cfg_dict["show_attributes"].append(att)
			else:
				if query["show_attributes"] not in cfg_dict["show_attributes"]:
					cfg_dict["show_attributes"].append(query["show_attributes"])
			del query["show_attributes"]
		
		#Check if there are attributes in query which are not in the "allowed" values
		query_keys = set(query.keys())
		invalid_keys = query_keys - valid_query_keys
		if len(invalid_keys) > 0:
			logger.error("Error in query {0}. Key(s) {1} are not known to UROPA.".format(i+1, invalid_keys))
			sys.exit()

		#Delete empty keys
		for key in list(query.keys()):
			if is_empty(query[key]):
				del query[key]

		# If filter_attribute is set, attribute_values has to be set as well
		if ("filter_attribute" in query and "attribute_values" not in query) or ("attribute_values" in query and "filter_attribute" not in query):
			logger.error("Error in query {0}. Keys for filter_attribute/attribute_values have to be set together.".format(i+1))
			sys.exit()

		##### Check content of individual keys #####
		if "distance" in query:
			try:
				if type(query["distance"]) == list:
					query["distance"] =  [int(query["distance"][0]), int(query["distance"][-1])]
				else:
					query["distance"] = [int(query["distance"]), int(query["distance"])]	#same distance in both positions
			except:
				logger.error("Error trying to convert distance \"{0}\" to integer - please check the format of this value.".format(query["distance"]))
				sys.exit()
		else:
			logger.error("No 'distance' not given in query: {0}".format(query))
			sys.exit()

		#Internals
		internalval = query.get("internals", False)
		query["internals"] = float(convert_values.get(str(internalval).lower(), internalval))	#if internals is a float, the value will not be converted
		if query["internals"] == 0:
			del query["internals"]

		#Keys which should be lists (sets) to search in 
		for key in ["feature", "feature_anchor", "attribute_values", "relative_location"]:
			if type(query.get(key, [])) != list:
				query[key] = set([query[key]])

		#Check the default values of input
		if "strand" in query:
			query["strand"] = query["strand"].lower()
			valid = set(['+', '-', 'same', 'opposite', 'ignore'])
			if query["strand"] not in valid:
				logger.error("Invalid strand ({0}) set in query {1}. Valid options are: {2}".format(query["strand"], i+1, valid))
				sys.exit()

		if "feature_anchor" in query:
			query["feature_anchor"] = set([element.lower() for element in query["feature_anchor"]])
			valid = set(["start", "center", "end"])
			invalid = query["feature_anchor"] - valid
			if len(invalid) > 0:
				logger.error("Invalid feature_anchor ({0}) set in query {1}. Valid options are: {2}".format(invalid, i+1, valid))
				sys.exit()

		if "relative_location" in query:
			query["relative_location"] = set([case_conversion.get(str(element).lower(), element) for element in query["relative_location"]])
			valid = set(["PeakInsideFeature", "FeatureInsidePeak", "Upstream", "Downstream", "OverlapStart", "OverlapEnd"])
			invalid = query["relative_location"] - valid
			if len(invalid) > 0:
				logger.error("Invalid relative_location ({0}) set in query {1}. Valid options are: {2}".format(invalid, i+1, valid))
				sys.exit()

		#Name the query if it was not already named
		if "name" not in query:
			query["name"] = query_prefix + str(i + 1)

	############## Warnings related to config/queries #############
	#Check for general input in queries, such as no show_attributes
	if len(cfg_dict.get("show_attributes", [])) == 0:
		logger.warning("No show_attributes given in config - no attributes for annotations are displayed in output.")

	#Catch duplicates in query names	
	query_names = [query["name"] for query in cfg_dict["queries"]]
	if len(query_names) != len(set(query_names)):
		logger.warning("Duplicated query names present: {0}".format(query_names))

	return(cfg_dict)



def config_string(cfg_dict):
	""" Pretty-print cfg_dict with one-line queries """

	upper_level = ["queries", "show_attributes", "priority",  "gtf", "bed", "prefix", "outdir", "threads", "output_by_query"]
	query_level = ["feature", "feature_anchor", "distance", "strand", "relative_location", "filter_attribute", "attribute_values", "internals", "name"]

	upper_lines = []
	for upper_key in upper_level:
		if upper_key == "queries":
			query_lines = "\"queries\":[\n"
			
			#Convert sets to lists
			for query in cfg_dict["queries"]:
				for key in query:
					if type(query[key]) == set:
						query[key] = list(query[key])

			query_strings = [json.dumps(query, sort_keys=True) for query in cfg_dict["queries"]]
			query_lines += "    " + ",\n    ".join(query_strings) + "\n          ]"

			upper_lines.append(query_lines)
		
		elif upper_key == "show_attributes" and upper_key in cfg_dict:
			upper_lines.append("\"{0}\": {1}".format(upper_key, json.dumps(cfg_dict[upper_key])))

		else:
			if upper_key in cfg_dict:
				upper_lines.append("\"{0}\": \"{1}\"".format(upper_key, cfg_dict[upper_key]))
		
	config_string = "{\n" + ",\n".join(upper_lines) + "\n}\n"

	return(config_string)

def check_file_access(fil, logger):
	""" Check if file exists and can be read """

	if os.path.isfile(fil):
		if not os.access(fil, os.R_OK):
			logger.error("File %s could not be read.", fil)
			sys.exit()
	else:
		logger.error("File %s does not exist.", fil)
		sys.exit()

def check_bed_format(bedfile, logger):
	""" Checks whether bedfile has the correct format """

	with open(bedfile) as f:
		for i, line in enumerate(f):
			if not re.match("(\S)+\s+([0-9]+)\s+([0-9]+)(\s+\S+)?(\s+[0-9,.]+)?(\s+[.\-+])?", line):
				logger.error("Line {0} in {1} is not proper bed format: {2}".format(i+1, bedfile, line))
				sys.exit()

	#todo: also check the number of columns in file

def parse_bedlines(bedlines, gtf_has_chr):
	""" Parse a list of bedlines to internal peaks format """

	peaks = [{}]*len(bedlines)	#initialize to length of bedlines; eliminates the need to append
	for i, line in enumerate(bedlines):
		columns = line.rstrip().split()

		#bed6-columns
		chrom, start, end = columns[0], int(columns[1]), int(columns[2])
		name = columns[3] if len(columns) > 3 else "peak_{0}".format(i+1) 
		score = columns[4] if len(columns) > 4 else "."
		strand = columns[5] if len(columns) > 5 else "."

		#Additional columns
		if len(columns) > 6:
			additional = columns[6:]
			additional_header = ["custom_" + str(val) for val in range(1,len(additional)+1)]
		else:
			additional_header = []
			additional = []

		#Key for the matching gtf chromosome
		if gtf_has_chr == True:
			gtf_chr = "chr" + chrom if not chrom.startswith("chr") else chrom		#add chr to match gtf if needed
		else:
			gtf_chr = chrom.replace("chr", "") #gtf chrom should not have chr-prefix
		
		peak_dict = {"gtf_chr": gtf_chr, "peak_chr":chrom, "peak_start":start, "peak_end":end, "peak_id":name, "peak_score":score, "peak_strand":strand}
		peak_dict.update(dict(zip(additional_header, additional)))
		peaks[i] = peak_dict

	return(peaks)


def check_chr(file, lines=100):
	"""Checks if a file has lines starting with 'chr'."""
    
	chrom = False
	count = 0
	with open(file, "r") as f:
		for line in f:
			if not line.startswith("#"):

				if line.startswith("chr"):
					chrom = True
					break

				count += 1	
				if count == lines:
					break

	return(chrom)

def subset_gtf(gtf, features, sub_gtf):
	"""Removes lines with features not in features from a gtf file."""
	feat2cut = set(np.unique(features))

	with open(gtf, "r") as f:
		lines = f.readlines()
		gtf_query_feat = [line for line in lines if not line.startswith("#") and line.split("\t")[2] in feat2cut]

	# List of lines containing only selected features from query
	with open(sub_gtf, "w") as f:
		f.write("".join(gtf_query_feat))	

def is_empty(value):
	""" Check whether value is empty """
	if (type(value) == str and value.strip() == "") or (type(value) == list and len(value) == 0):
		return(True)
	else:
		return(False)

def sorted_file_writer(q, file_dict, logger_options):
	""" Write strings in the correct order coming from Queue """

	#Start logger
	logger = UROPALogger(**logger_options)

	#Open handles for files
	file_handles = {}
	for key in file_dict:
		file = file_dict[key]
		try:
			file_handles[key] = open(file, "w")
		except Exception as e:
			print("Error opening file {0} in file_writer".format(file))
			print(e)
			return(0)

	#Fetching string content from queue
	idx_to_write = {key: 0 for key in file_dict}	#Start with idx == 0
	ready_to_write = {key:{} for key in file_dict}

	delta = datetime.timedelta(seconds=10) #Write debug status every 10 seconds
	prev_time = datetime.datetime.now()
	while True:

		#Write debug status
		current_time = datetime.datetime.now()
		if current_time - prev_time > delta: 
			n_ready_to_write = {key: sorted(list(ready_to_write[key].keys())) for key in ready_to_write}	#dict of idx-lists per key
			logger.debug("Writer task status | ready to write: {0}".format(n_ready_to_write))
			prev_time = current_time

		#Fetch annotated sites from queue
		try:
			(key, idx, content) = q.get() #idx is an integer starting at 0, content is a string

			if key == None:
				break	#no more content to write 
			
			#Add content to indexes ready to write
			ready_to_write[key][idx] = content

			#Write across keys
			for key in file_dict:
				while idx_to_write[key] in ready_to_write[key]:
					content = ready_to_write[key][idx_to_write[key]]
					file_handles[key].write(content)

					del ready_to_write[key][idx_to_write[key]] #Content has been written for this idx; free up memory!
					idx_to_write[key] += 1	#increment to write next block
					

		except Exception as e:
			logger.error('Error occurred during file writing: {0}'.format(type(e)))
			logger.debug("Full traceback: {0}".format(traceback.format_exc()))
			return(1)

	#Got all content in queue, close file
	for key in file_dict:
		file_handles[key].close()
	
	return(0)	


import os
import sys
import re
import setuptools
from setuptools import setup

#Requires minimum python 3.2
if sys.version_info[0] > 3 or (sys.version_info[0] == 3 and sys.version_info[1] >= 2):
  pass
else:
    sys.exit("ERROR: UROPA install requires python>=3.2")

#Path of setup file to establish version
setupdir = os.path.abspath(os.path.dirname(__file__))

def find_version(init_file):
  version_file = open(init_file).read()
  version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
  if version_match:
    return version_match.group(1)
  else:
    raise RuntimeError("Unable to find version string.")

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='uropa',
      version=find_version(os.path.join(setupdir, "uropa", "__init__.py")),
      description='UROPA is a command line based tool, intended for genomic region annotation',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.molgen.mpg.de/loosolab/UROPA',
      author='Jens Preussner',
      author_email='jens.preussner@mpi-bn.mpg.de',
      license='MIT',
      packages=['uropa'],
      entry_points = {
        'console_scripts': ['uropa = uropa.uropa:main']
      },
      scripts = ['utils/uropa_summary.R','utils/uropa2gtf.R'],
      python_requires='>=3.2',
      install_requires=[
        'pysam',
        'psutil',
        'numpy'
      ],
      classifiers = [
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python '
      ],
      zip_safe=False,
      include_package_data=True)
